from zoneinfo import ZoneInfo
from scrapy import Spider
from scrapy.exceptions import IgnoreRequest
from datetime import datetime, timedelta
import dateparser




class PreProcessingMiddleware:
    """
    Middleware to preprocess requests before forwarding.
    Handles deduplication
    """

    def process_request(self, request, spider: Spider) -> None:
        _dt = request.meta.pop("_dt", None)
        _dt_format = request.meta.pop("_dt_format", None)
        if _dt:
            if not self.is_recent(_dt, _dt_format, request.url, spider):
                raise IgnoreRequest
        return None
    
    def is_recent(self, date_str: str, date_format: str, debug_info: str, spider: Spider) -> bool:
        """
        Check if the date is recent (within the last 2 days).
        """
        try:
            if not date_str:
                return True
            utc_today = datetime.now(ZoneInfo('UTC')).date()
            input_date = dateparser.parse(date_string=date_str, date_formats=[date_format] if date_format is not None else None).date()
            return input_date >= (utc_today - timedelta(days=2))
        except Exception as e:
            spider.logger.error(f"{str(e)}: {debug_info} ")
            return False



from scrapy.crawler import Crawler
from scrapy.statscollectors import StatsCollector
from typing import Self
from scrapy import Spider, signals
from scrapy.http import Request, Response
from time import time



class ZenExtension:
    """
    Allows to calculate average latency across requests
    """

    def __init__(self, stats: StatsCollector) -> None:
        self.stats = stats
        self.response_count = 0
        self.total_latency = 0.0
        self.max_latency = None
        self.min_latency = None

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        ext = cls(crawler.stats)
        crawler.signals.connect(ext.response_received, signal=signals.response_received)
        crawler.signals.connect(ext.request_reached_downloader, signal=signals.request_reached_downloader)
        crawler.signals.connect(ext.spider_closed, signal=signals.spider_closed)
        return ext
    
    def request_reached_downloader(self, request: Request, spider: Spider) -> None:
        request.meta['zen_start_time'] = time()
    
    def response_received(self, response: Response, request: Request, spider: Spider) -> None:
        download_latency = time() - request.meta.pop("zen_start_time")
        if self.min_latency is None:
            self.min_latency = download_latency
            self.max_latency = download_latency
        else:
            if download_latency > self.max_latency:
                self.max_latency = download_latency
            if download_latency < self.min_latency:
                self.min_latency = download_latency
        self.total_latency += download_latency
        self.response_count += 1

    def spider_closed(self, spider: Spider, reason: str) -> None:
        if self.response_count > 0:
            avg_response_latency = self.total_latency / self.response_count
            self.stats.set_value("zen/avg_latency_seconds", f"{avg_response_latency:.2f}")
            self.stats.set_value("zen/min_latency_seconds", f"{self.min_latency:.2f}")
            self.stats.set_value("zen/max_latency_seconds", f"{self.max_latency:.2f}")
            self.stats.set_value("zen/response_count", self.response_count) 



import logging
from pathlib import Path
import time
from scrapy.dupefilters import RFPDupeFilter
from scrapy.http import Request
from scrapy.utils.request import (
    RequestFingerprinter,
    RequestFingerprinterProtocol,
)



class ZenDupeFilter(RFPDupeFilter):

    def __init__(
        self,
        path: str | None = None,
        debug: bool = False,
        *,
        fingerprinter: RequestFingerprinterProtocol | None = None,
    ) -> None:
        self.file = None
        self.fingerprinter: RequestFingerprinterProtocol = (
            fingerprinter or RequestFingerprinter()
        )
        self.fingerprints: set[str] = set()
        self.logdupes = True
        self.debug = debug
        self.logger = logging.getLogger(__name__)
        if path:
            self.file = Path(path, "requests.seen").open("a+", encoding="utf-8")
            self.file.seek(0)
            valid_entries = set()
            for x in self.file:
                fp, timestamp = x.rstrip().split("_")
                if self.is_timestamp_older_than(int(timestamp), 7):
                    continue
                self.fingerprints.add(fp)
                valid_entries.add(x.rstrip())
            
            self.logger.info(f"Loaded {len(valid_entries)} entries from {path}")
            self.file.truncate(0) # clear
            self.file.seek(0) # move to start
            for entry in valid_entries:
                self.file.write(entry + "\n")
            self.file.flush()
            

    def request_seen(self, request: Request) -> bool:
        fp = self.request_fingerprint(request)
        if fp in self.fingerprints:
            return True
        self.fingerprints.add(fp)
        if self.file:
            fp = f"{fp}_{int(time.time())}"
            self.file.write(fp + "\n")
        return False


    @staticmethod
    def is_timestamp_older_than(timestamp: int, days: int) -> bool:
        current_time = int(time.time())
        days_in_seconds = days * 24 * 60 * 60
        return (current_time - timestamp) > days_in_seconds